import torch
import torch.nn.functional as F
from math import exp
import numpy as np

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

def gaussian(window_size, sigma):
    gauss = torch.Tensor([exp(-(x - window_size//2)**2/float(2*sigma**2)) for x in range(window_size)])
    return gauss/gauss.sum()


def create_window(window_size, channel=1):
    _1D_window = gaussian(window_size, 1.5).unsqueeze(1)
    _2D_window = _1D_window.mm(_1D_window.t()).float().unsqueeze(0).unsqueeze(0).to(device)
    window = _2D_window.expand(channel, 1, window_size, window_size).contiguous()
    return window

def create_window_3d(window_size, channel=1):
    _1D_window = gaussian(window_size, 1.5).unsqueeze(1)
    _2D_window = _1D_window.mm(_1D_window.t())
    _3D_window = _2D_window.unsqueeze(2) @ (_1D_window.t())
    window = _3D_window.expand(1, channel, window_size, window_size, window_size).contiguous().to(device)
    return window


def ssim(img1, img2, window_size=11, window=None, size_average=True, full=False, val_range=None):
    # Value range can be different from 255. Other common ranges are 1 (sigmoid) and 2 (tanh).
    if val_range is None:
        if torch.max(img1) > 128:
            max_val = 255
        else:
            max_val = 1

        if torch.min(img1) < -0.5:
            min_val = -1
        else:
            min_val = 0
        L = max_val - min_val
    else:
        L = val_range

    padd = 0
    (_, channel, height, width) = img1.size()
    if window is None:
        real_size = min(window_size, height, width)
        window = create_window(real_size, channel=channel).to(img1.device)
    
    # mu1 = F.conv2d(img1, window, padding=padd, groups=channel)
    # mu2 = F.conv2d(img2, window, padding=padd, groups=channel)
    mu1 = F.conv2d(F.pad(img1, (5, 5, 5, 5), mode='replicate'), window, padding=padd, groups=channel)
    mu2 = F.conv2d(F.pad(img2, (5, 5, 5, 5), mode='replicate'), window, padding=padd, groups=channel)

    mu1_sq = mu1.pow(2)
    mu2_sq = mu2.pow(2)
    mu1_mu2 = mu1 * mu2

    sigma1_sq = F.conv2d(F.pad(img1 * img1, (5, 5, 5, 5), 'replicate'), window, padding=padd, groups=channel) - mu1_sq
    sigma2_sq = F.conv2d(F.pad(img2 * img2, (5, 5, 5, 5), 'replicate'), window, padding=padd, groups=channel) - mu2_sq
    sigma12 = F.conv2d(F.pad(img1 * img2, (5, 5, 5, 5), 'replicate'), window, padding=padd, groups=channel) - mu1_mu2

    C1 = (0.01 * L) ** 2
    C2 = (0.03 * L) ** 2

    v1 = 2.0 * sigma12 + C2
    v2 = sigma1_sq + sigma2_sq + C2
    cs = torch.mean(v1 / v2)  # contrast sensitivity

    ssim_map = ((2 * mu1_mu2 + C1) * v1) / ((mu1_sq + mu2_sq + C1) * v2)

    if size_average:
        ret = ssim_map.mean()
    else:
        ret = ssim_map.mean(1).mean(1).mean(1)

    if full:
        return ret, cs
    return ret


def ssim_matlab(img1, img2, window_size=11, window=None, size_average=True, full=False, val_range=None):
    # Value range can be different from 255. Other common ranges are 1 (sigmoid) and 2 (tanh).
    if val_range is None:
        if torch.max(img1) > 128:
            max_val = 255
        else:
            max_val = 1

        if torch.min(img1) < -0.5:
            min_val = -1
        else:
            min_val = 0
        L = max_val - min_val
    else:
        L = val_range

    padd = 0
    (_, _, height, width) = img1.size()
    if window is None:
        real_size = min(window_size, height, width)
        window = create_window_3d(real_size, channel=1).to(img1.device)
        # Channel is set to 1 since we consider color images as volumetric images

    img1 = img1.unsqueeze(1)
    img2 = img2.unsqueeze(1)

    mu1 = F.conv3d(F.pad(img1, (5, 5, 5, 5, 5, 5), mode='replicate'), window, padding=padd, groups=1)
    mu2 = F.conv3d(F.pad(img2, (5, 5, 5, 5, 5, 5), mode='replicate'), window, padding=padd, groups=1)

    mu1_sq = mu1.pow(2)
    mu2_sq = mu2.pow(2)
    mu1_mu2 = mu1 * mu2

    sigma1_sq = F.conv3d(F.pad(img1 * img1, (5, 5, 5, 5, 5, 5), 'replicate'), window, padding=padd, groups=1) - mu1_sq
    sigma2_sq = F.conv3d(F.pad(img2 * img2, (5, 5, 5, 5, 5, 5), 'replicate'), window, padding=padd, groups=1) - mu2_sq
    sigma12 = F.conv3d(F.pad(img1 * img2, (5, 5, 5, 5, 5, 5), 'replicate'), window, padding=padd, groups=1) - mu1_mu2

    C1 = (0.01 * L) ** 2
    C2 = (0.03 * L) ** 2

    v1 = 2.0 * sigma12 + C2
    v2 = sigma1_sq + sigma2_sq + C2
    cs = torch.mean(v1 / v2)  # contrast sensitivity

    ssim_map = ((2 * mu1_mu2 + C1) * v1) / ((mu1_sq + mu2_sq + C1) * v2)

    if size_average:
        ret = ssim_map.mean()
    else:
        ret = ssim_map.mean(1).mean(1).mean(1)

    if full:
        return ret, cs
    return ret


def msssim(img1, img2, window_size=11, size_average=True, val_range=None, normalize=False):
    device = img1.device
    weights = torch.FloatTensor([0.0448, 0.2856, 0.3001, 0.2363, 0.1333]).to(device)
    levels = weights.size()[0]
    mssim = []
    mcs = []
    for _ in range(levels):
        sim, cs = ssim(img1, img2, window_size=window_size, size_average=size_average, full=True, val_range=val_range)
        mssim.append(sim)
        mcs.append(cs)

        img1 = F.avg_pool2d(img1, (2, 2))
        img2 = F.avg_pool2d(img2, (2, 2))

    mssim = torch.stack(mssim)
    mcs = torch.stack(mcs)

    # Normalize (to avoid NaNs during training unstable models, not compliant with original definition)
    if normalize:
        mssim = (mssim + 1) / 2
        mcs = (mcs + 1) / 2

    pow1 = mcs ** weights
    pow2 = mssim ** weights
    # From Matlab implementation https://ece.uwaterloo.ca/~z70wang/research/iwssim/
    output = torch.prod(pow1[:-1] * pow2[-1])
    return output


# Classes to re-use window
class SSIM(torch.nn.Module):
    def __init__(self, window_size=11, size_average=True, val_range=None):
        super(SSIM, self).__init__()
        self.window_size = window_size
        self.size_average = size_average
        self.val_range = val_range

        # Assume 3 channel for SSIM
        self.channel = 3
        self.window = create_window(window_size, channel=self.channel)

    def forward(self, img1, img2):
        (_, channel, _, _) = img1.size()

        if channel == self.channel and self.window.dtype == img1.dtype:
            window = self.window
        else:
            window = create_window(self.window_size, channel).to(img1.device).type(img1.dtype)
            self.window = window
            self.channel = channel

        _ssim = ssim(img1, img2, window=window, window_size=self.window_size, size_average=self.size_average)
        dssim = (1 - _ssim) / 2
        return dssim

class MSSSIM(torch.nn.Module):
    def __init__(self, window_size=11, size_average=True, channel=3):
        super(MSSSIM, self).__init__()
        self.window_size = window_size
        self.size_average = size_average
        self.channel = channel

    def forward(self, img1, img2):
        return msssim(img1, img2, window_size=self.window_size, size_average=self.size_average)



import patchify
import json
import numpy as np
from patchify import patchify, unpatchify
import skimage
from skimage.morphology import disk, binary_erosion, binary_dilation
import yeastvision.models as models
from os.path import join
import os
from yeastvision.utils import capitalize_first_letter

CUSTOM_MODELS_FILE = "model_types.json"
MODEL_DIR = models.__path__[0]
with open(os.path.join(MODEL_DIR, CUSTOM_MODELS_FILE), "r") as file:
    CUSTOM_MODEL_TYPES = json.load(file)

def is_RGB(ims):
    return ims.shape[-1] == 3

def getBuiltInModelTypes():
    '''Retrieves the builtin model names'''
    models = [model for model in os.listdir(MODEL_DIR) if os.path.isdir(join(MODEL_DIR, model)) and model != "__pycache__"]
    return models

def getModelsByType(model_type):
    '''Returns all the weights that exist within model_type directory'''
    model_dir = os.path.join(MODEL_DIR, model_type)
    return [file for file in os.listdir(model_dir) if "." not in file and file != "__pycache__"]

def getModelLoadedStatus(model):
    '''Determines whether or not a model is loaded. Built for default models type
    Assumes that model name is identical to model type'''
    return os.path.exists(produce_weight_path(model, model))

def produce_weight_path(modeltype, modelname):
    return join(MODEL_DIR, modeltype, modelname)

def shrink_bud(bud_mask, footprint = None, kernel_size = 2):
    if not footprint:
        footprint = disk(kernel_size)
    return binary_erosion(bud_mask, footprint)

def enlarge_bud(bud_mask, footprint = None, kernel_size = 4):
    if not footprint:
        footprint = disk(kernel_size)
    return binary_dilation(bud_mask, footprint)

def normalizeIm(im):
    return (im - np.min(im)) / (np.max(im) - np.min(im))

def addMasks(maskToAdd, origMask):
    newMask = origMask.copy()
    for i in np.unique(maskToAdd)[1:]:
        boolMask = maskToAdd==i
        valExists, maskValToReplace = getMaskOverlapVal(boolMask, origMask)
        
        if valExists:
            newMask[newMask==maskValToReplace]=0
            newMask[boolMask] = maskValToReplace
        else:
            newMask[boolMask] = maskValToReplace
    return newMask

def getMaskOverlapVal(boolMask, fullMask):
    vals, counts = np.unique(fullMask[boolMask], return_counts=True)
    if len(vals)>1:
        val = vals[np.argmax(counts)]
        if val==0:
            return False, fullMask.max()+1
        else:
            return True, val
    else:
        if vals[0]==0:
            return False, fullMask.max()+1
        else:
            return True, vals[0]
        
def test_time_aug(im, model):
    temp_test_img = im[0,:,:,0]
    p0 = model.predict(np.expand_dims(temp_test_img, axis=0), verbose = 0)[0][:, :, 0]

    p1 = model.predict(np.expand_dims(np.fliplr(temp_test_img), axis=0), verbose = 0)[0][:, :, 0]
    p1 = np.fliplr(p1)

    p2 = model.predict(np.expand_dims(np.flipud(temp_test_img), axis=0), verbose = 0)[0][:, :, 0]
    p2 = np.flipud(p2)

    p3 = model.predict(np.expand_dims(np.fliplr(np.flipud(temp_test_img)), axis=0), verbose =0 )[0][:, :, 0]
    p3 = np.fliplr(np.flipud(p3))

    p = ((p0 + p1 + p2 + p3)/4).astype(np.float32)
    return p

def prediction(models, image, input_shape = 256, buffer_size = 10,  double_patch = False, do_thresh = True, thresh = 0.50, pad_mode = "symmetric", preprocess = skimage.exposure.equalize_adapthist, do_aug = True):
    '''
    Params
    ---------
    Image (ndarray) -  2D image
    Model (tf.Model) - trained unet
    patch_shape (tuple) - size of 2D tiles after padding, should match model input reqs
    buffer_size (int) - size of additional paddding added
    
    Returns
    --------
    Ndarray uint8
    '''
    if type(models) != list:
        models = [models]

    patch_size = input_shape - 2*buffer_size
    if double_patch:
        step_size = patch_size//2    
    else:
        step_size = patch_size

    large_image = pad_for_patching(image, patch_size = patch_size)
    patches = patchify(large_image, (patch_size, patch_size), step=step_size) 
    predicted_patches = []
    for i in range(patches.shape[0]):
        for j in range(patches.shape[1]):
            single_patch = patches[i,j,:,:]

            # we pad our patch for prediction
            #single_patch = np.pad(single_patch, ((10,10), (10,10)), mode = pad_mode)
            single_patch = np.pad(single_patch, ((buffer_size,buffer_size), (buffer_size,buffer_size)), mode = pad_mode)
            # add a leading and trailing dimension 
            single_patch_norm = np.expand_dims(preprocess(np.array(single_patch)),2)
            single_patch_input=np.expand_dims(single_patch_norm, 0)
            
            avg_image = np.zeros((input_shape, input_shape), dtype = np.float32)
            for model in models:
                if do_aug:
                    single_patch_prediction = test_time_aug(single_patch_input, model)
                else:
                    single_patch_prediction = (model.predict(single_patch_input, verbose = 0)[0,:,:,0])
                avg_image = np.add(avg_image, single_patch_prediction) 
            avg_image = avg_image/len(models)
            
            if do_thresh:
                single_patch_prediction = (avg_image>thresh).astype(np.uint8)

            # crop additional padding, reducing image back to size patch_size
            single_patch_prediction = single_patch_prediction[buffer_size:(input_shape-buffer_size), buffer_size:(input_shape-buffer_size)]
            predicted_patches.append(single_patch_prediction)

    # unpatchify the list of images, using format required by patchify module
    # reshape 
    predicted_patches = np.array(predicted_patches)
    predicted_patches_reshaped = np.reshape(predicted_patches, (patches.shape[0], patches.shape[1], patch_size,patch_size) )
    reconstructed_large_image = unpatchify(predicted_patches_reshaped, large_image.shape)

    # remove padding added to large image during pad_for_patching()
    reconstructed_image = reconstructed_large_image[0:image.shape[0], 0:image.shape[1]]
    
    return reconstructed_image

def pad_for_patching(image, patch_size = 256, pad_mode = "constant"):
    '''adds padding determined by get_pad_dim to image and returns a padded copy of same type '''

    def get_closest_int(num, target = 256):
        '''returns closest integer greater or equal to num divisible by target '''
        return (num + target) - (num % target)

    def get_pad_dim_for_patching(shape, patch_size = 256):
        '''returns the neccesary additions vertically and horizonatally for an image to be divisible
        evenly by the specified patch_size (patches must be symmetrical)'''
        r,c = shape
        r_pad = (get_closest_int(r, target = patch_size) - r)
        c_pad = (get_closest_int(c, target = patch_size) - c)
        return r_pad, c_pad

    
    r_pad, c_pad = get_pad_dim_for_patching(image.shape[:2], patch_size = patch_size) # 2D images only
    padded_image = np.pad(image, ((0, r_pad), (0, c_pad)), mode = pad_mode)
    return padded_image


def patchify_for_train(ims, input_shape = 256, double_patch = True):

    patch_size = input_shape 
    if double_patch:
        step_size = patch_size//2    
    else:
        step_size = patch_size

    out_patches = []
    for image in ims:
        large_image = pad_for_patching(image, patch_size = patch_size)
        patches = patchify(large_image, (patch_size, patch_size), step=step_size) 
        for i in range(patches.shape[0]):
            for j in range(patches.shape[1]):
                out_patches.append(patches[i,j,:,:])
    return out_patches


import numpy as np
from skimage.draw import disk
from skimage.feature import blob_log as bl
from skimage.exposure import equalize_adapthist
import math


def normalize_im(im, clip = True):
    """
    Normalizes a given image such that the values range between 0 and 1.     
    
    Parameters
    ---------- 
    im : 2d-array
        Image to be normalized.
    clip: boolean
        Whether or not to set im values <0 to 0
        
    Returns
    -------
    im_norm: 2d-array
        Normalized image ranging from 0.0 to 1.0. Note that this is now
        a floating point image and not an unsigned integer array. 
    """ 
    assert not np.any(np.isnan(im))
    if clip:
        im[im<0] = 0
    if im.max()==0:
        return im.astype(np.float32)
    im_norm = (im - im.min()) / (im.max() - im.min())
    return im_norm.astype(np.float32)


def draw_blobs(ims, min_size, max_size, threshold):
    masks = np.zeros_like(ims)
    for idx, im in enumerate(ims):
        im = equalize_adapthist(im)
        blobs = bl(ims, min_sigma = min_size, max_sigma = max_size, threshold = threshold)
        blobs[:, 2] = blobs[:, 2] * math.sqrt(2)

        label = 1
        for blob in blobs:
            r,c,radius = blob
            rr, cc = disk((int(r),int(c)), int(radius), shape = im.shape)
            masks[idx, rr, cc] =  label
            label+=1
    return masks




from skimage.measure import regionprops
import numpy as np
import cv2

def normalize_dict_by_sum(data):
    total = sum(data.values())  # Use the built-in sum() function for efficiency
    if total == 0:
        return {k: 0 for k in data}  # Avoid division by zero by returning zero for all keys
    return {k: v / total for k, v in data.items()}  # Use a dictionary comprehension for normalization


def get_cell_bbox(array, cellval, desired_shape):
    # Get the center coordinates
    centroid = regionprops((array==cellval).astype(np.uint8))[0].centroid
    r,c = centroid
    r,c = round(r), round(c)
    center_row, center_col = r,c

    # Get the desired shape dimensions
    desired_height, desired_width = desired_shape

    # Calculate the starting and ending row indices
    start_row = center_row - desired_height // 2
    end_row = start_row + desired_height

    # Calculate the starting and ending column indices
    start_col = center_col - desired_width // 2
    end_col = start_col + desired_width

    # Check if the indices are out of bounds
    if start_row < 0:
        start_row = 0
        end_row = min(start_row + desired_height, array.shape[0])
    elif end_row > array.shape[0]:
        end_row = array.shape[0]
        start_row = max(end_row - desired_height, 0)

    if start_col < 0:
        start_col = 0
        end_col = min(start_col + desired_width, array.shape[1])
    elif end_col > array.shape[1]:
        end_col = array.shape[1]
        start_col = max(end_col - desired_width, 0)

    # Get the actual cropped region from the array
    cropped_array = array[start_row:end_row, start_col:end_col]

    return cropped_array

def get_bbox_coords(bool_im, cell_margin = 20):
    centroid = regionprops(bool_im.astype(np.uint8))[0].centroid
    r_size, c_size = bool_im.shape
    r,c = centroid
    r,c = round(r), round(c)
    bbox = slice(max([0, r-cell_margin]), min([r_size, r+cell_margin])), slice(max([0, c-cell_margin]), min([c_size, c+cell_margin]))
    return bbox

def avg_cell_size(Matmasks,nu,no_obj):
    avg = [[[] for _ in range(nu)] for _ in range(no_obj.astype(int)+1)]
    avg_cell = [0]*no_obj.astype(int)

    for iv in range(1,no_obj.astype(int)+1):
        for its in range(nu):
            avg[iv-1][its] = np.average(Matmasks[0][its]==iv)
        
    for iv in range(1,no_obj.astype(int)+1):
        count = 0
        for its in range(nu):
            if avg[iv-1][its] != 0:
                count = count + 1
        avg_cell[iv-1] = np.sum(avg[iv-1])/count
        
    return np.min(avg_cell), np.average(avg_cell), np.max(avg_cell)

def get_wind_coord1(I3, cell_margin=20):
    
    # Get the dimensions of the input image
    x_size, y_size = I3.shape
    # If the input image contains non-zero pixels, find the bounding box of the wind object
    if np.sum(I3) > 0:
        # Threshold the image to create a binary image
        _, binary_image = cv2.threshold(I3, 0, 255, cv2.THRESH_BINARY + cv2.THRESH_OTSU)
        
        # Find the contours in the binary image
        contours, _ = cv2.findContours(binary_image, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
        # Loop over each contour to compute the centroid and bounding box
        s_f = []
        for contour in contours:
            # Compute the centroid
            M = cv2.moments(contour)
            cx = int(M['m10'] / (M['m00'] + 1e-6))  # Add small epsilon to avoid division by zero
            cy = int(M['m01'] / (M['m00'] + 1e-6)) 
            # Get the bounding box coordinates
            x, y, w, h = cv2.boundingRect(contour)
            # Save the centroid and bounding box information
            s_f.append({'Centroid': (cx, cy), 'BoundingBox': (x, y, w, h)})
        # If there is at least one wind object, get the bounding box coordinates and margins
        if s_f:
            bbox = np.round(s_f[0]['BoundingBox'])
            lower_x_limit = max(1, bbox[0]-cell_margin)
            upper_x_limit = min(y_size, bbox[0]+bbox[2]+cell_margin)
            lower_y_limit = max(1, bbox[1]-cell_margin)
            upper_y_limit = min(x_size, bbox[1]+bbox[3]+cell_margin)
            if upper_x_limit >= I3.shape[1]:
                upper_x_limit = I3.shape[1]-1
            if upper_y_limit >= I3.shape[0]:
                upper_y_limit = I3.shape[0]-1
            x_cn = range(lower_x_limit, upper_x_limit+1)
            y_cn = range(lower_y_limit, upper_y_limit+1)
        # If no wind object is found, raise an error
        else:
            print('empty or multiple object given - error')
    # If the input image is all zero pixels, use the entire image as the bounding box
    else:
        x_cn = np.arange(1, y_size+1)
        y_cn = np.arange(1, x_size+1)

    return x_cn, y_cn


from yeastvision.models.model import Model as CustomModel
import cv2
import numpy as np
from cellpose.transforms import normalize99
import torch
from cellpose.models import CellposeModel, size_model_path, SizeModel, Cellpose
from cellpose.core import assign_device

from yeastvision.utils import normalize_im


class CellposeAllowPreTrainedModel(Cellpose):
    """ main model which combines SizeModel and CellposeModel, and allows for a pretrained model to be loaded
    for cellpose model

    Parameters
    ----------

    gpu: bool (optional, default False)
        whether or not to use GPU, will check if GPU available

    model_type: str (optional, default 'cyto')
        'cyto'=cytoplasm model; 'nuclei'=nucleus model; 'cyto2'=cytoplasm model with additional user images

    net_avg: bool (optional, default False)
        loads the 4 built-in networks and averages them if True, loads one network if False

    device: torch device (optional, default None)
        device used for model running / training 
        (torch.device('cuda') or torch.device('cpu')), overrides gpu input,
        recommended if you want to use a specific GPU (e.g. torch.device('cuda:1'))
    
    pretrained_cp_model: path to custom cellpose weights (optional, default None)
        weights for a pretrained CellposeModel model

    """
    def __init__(self, gpu=False, model_type='cyto', net_avg=False, device=None, pretrained_model = None):
        super(Cellpose, self).__init__()
        self.torch = True
        
        # assign device (GPU or CPU)
        sdevice, gpu = assign_device(self.torch, gpu)
        self.device = device if device is not None else sdevice
        self.gpu = gpu
        
        model_type = 'cyto' if model_type is None else model_type
        
        self.diam_mean = 30. #default for any cyto model 
        nuclear = 'nuclei' in model_type
        if nuclear:
            self.diam_mean = 17. 
        
        self.cp = CellposeModel(device=self.device, gpu=self.gpu,
                                model_type=model_type,
                                diam_mean=self.diam_mean,
                                net_avg=net_avg,
                                pretrained_model=pretrained_model)
        self.cp.model_type = model_type
        
        # size model not used for bacterial model
        self.pretrained_size = size_model_path(model_type, self.torch)
        self.sz = SizeModel(device=self.device, pretrained_size=self.pretrained_size,
                            cp_model=self.cp)
        self.sz.model_type = model_type


class CustomCPWrapper(CustomModel):
    hyperparams  = {  
    "mean_diameter":-1.0, 
    "flow_threshold":0.4, 
    "cell_probability_threshold": 0}
    types = [float, float,float]
    
    loss = "categorical_crossentropy"
    trainparams = {
                "learning_rate":0.1,
                "weight_decay": 0.0001,
                "n_epochs":100
                }

    def __init__(self, params, weights):
        super().__init__(params, weights)

        self.mean_diam = self.params["mean_diameter"]
        self.do_size_estimation = self.mean_diam == -1
        self.no_diam = self.mean_diam == 0
        self.use_size_model = self.do_size_estimation

        if self.use_size_model:
            if self.do_size_estimation:
                self.mean_diam = None
            self.model  = CellposeAllowPreTrainedModel(gpu=True, pretrained_model=self.weights)
        else:
            self.model = CellposeModel(gpu=True, pretrained_model=self.weights)
            if self.no_diam:
                self.mean_diam = self.model.diam_labels
    
    def eval_params(self, params):
        a = {"diameter":self.mean_diam,
                "channels": [0,0],
                "cellprob_threshold": params["cell_probability_threshold"],
                "flow_threshold": params["flow_threshold"],
                "do_3D": False,
                "min_size":-1}
        return a
    
    def process_probability(self, rawProb):
        return (np.clip(normalize99(rawProb), 0, 1) * 255).astype(np.uint8)

    def train(self, ims, labels, params, savepath):
        ims = [cv2.merge((im,im,im)) for im in ims]
        
        modelToTrain = self.model if isinstance(self.model, CellposeModel) else self.model.cp
        modelToTrain.train(ims, labels, 
                                        channels=[0,0], 
                                        save_path=params["dir"], 
                                        min_train_masks=1,
                                        nimg_per_epoch=8,
                                        learning_rate = params['learning_rate'], 
                                        weight_decay = params['weight_decay'], 
                                        n_epochs = int(params['n_epochs']),
                                        model_name = params["model_name"])
    def get_masks_and_flows(self, ims):
        if self.use_size_model:
            masks, flows, _, _ = self.model.eval(ims, **self.eval_params(self.params))
        else:
            masks, flows, _ = self.model.eval(ims, **self.eval_params(self.params))
        return masks,flows
    
    def process_flows(self, flow_list) -> np.ndarray[np.uint8]:
        '''Reduces cellpose 3D flows into 2D images by taking the magnitude of the flow vector'''
        # Calculate the magnitude of the vector along the last dimension
        flowsXY = np.array([normalize_im(np.linalg.norm(flow[0], axis=-1)) for flow in flow_list], dtype=np.float32) * 255
        return flowsXY.astype(np.uint8)

    @classmethod
    @torch.no_grad()
    def run(cls, ims, params, weights):
        '''Class method for running a built-in model type
        
        Returns
        - masks (np.ndarray[uint16]): the segmentation results where each pixel is given an index according to the object
        - cellprobs (np.ndarray[uint8]): the probability of each pixel being an object or background. Scaled such that 255 is 100% confidence
                                        and 0 is 0% confidence
        - flowsXY (np.ndarray[np.uint8]): each pixel gives the magnitude of the outputted flow vector at that location
        '''
        params = params if params else cls.hyperparams
        model = cls(params, weights)
        #ims3D = [cv2.merge((im,im,im)) for im in ims]
        if isinstance(ims, np.ndarray):
            ims = list(ims)
        masks, flows = model.get_masks_and_flows(ims)
        cellprobs = [flow[2] for flow in flows]
        flowsXY = model.process_flows(flows)
        cellprobs = np.array((model.process_probability(cellprobs)), dtype = np.uint8)
        masks = np.array(masks, dtype = np.uint16)
        del model
        return masks, cellprobs, flowsXY